@@ -29,17 +29,21 @@ from keystoneclient.auth.identity import v3 as v3_auth
 from keystoneclient import discover
 from keystoneclient.openstack.common.apiclient import exceptions as ks_exc
 from keystoneclient import session as kssession
+from oslo_utils import encodeutils
+import six
 import six.moves.urllib.parse as urlparse
 
 
 import ironicclient
 from ironicclient import client as iroclient
+from ironicclient.common import http
+from ironicclient.common.i18n import _
 from ironicclient.common import utils
 from ironicclient import exc
 from ironicclient.openstack.common import cliutils
-from ironicclient.openstack.common import gettextutils
 
-gettextutils.install('ironicclient')
+
+LATEST_API_VERSION = ('1', 'latest')
 
 
 class IronicShell(object):
@@ -186,9 +190,10 @@ class IronicShell(object):
 
         parser.add_argument('--ironic-api-version',
                             default=cliutils.env(
-                            'IRONIC_API_VERSION', default='1'),
-                            help='Defaults to env[IRONIC_API_VERSION] '
-                            'or 1')
+                                'IRONIC_API_VERSION', default='1'),
+                            help='Accepts 1.x (where "x" is microversion) '
+                                 'or "latest", Defaults to '
+                                 'env[IRONIC_API_VERSION] or 1')
 
         parser.add_argument('--ironic_api_version',
                             help=argparse.SUPPRESS)
@@ -201,6 +206,16 @@ class IronicShell(object):
         parser.add_argument('--os_service_type',
                             help=argparse.SUPPRESS)
 
+        parser.add_argument('--os-endpoint',
+                            default=cliutils.env('OS_SERVICE_ENDPOINT'),
+                            help='Specify an endpoint to use instead of '
+                                 'retrieving one from the service catalog '
+                                 '(via authentication). '
+                                 'Defaults to env[OS_SERVICE_ENDPOINT].')
+
+        parser.add_argument('--os_endpoint',
+                            help=argparse.SUPPRESS)
+
         parser.add_argument('--os-endpoint-type',
                             default=cliutils.env('OS_ENDPOINT_TYPE'),
                             help='Defaults to env[OS_ENDPOINT_TYPE] or '
@@ -209,6 +224,25 @@ class IronicShell(object):
         parser.add_argument('--os_endpoint_type',
                             help=argparse.SUPPRESS)
 
+        parser.add_argument('--max-retries', type=int,
+                            help='Maximum number of retries in case of '
+                            'conflict error (HTTP 409). '
+                            'Defaults to env[IRONIC_MAX_RETRIES] or %d. '
+                            'Use 0 to disable retrying.'
+                            % http.DEFAULT_MAX_RETRIES,
+                            default=cliutils.env(
+                                'IRONIC_MAX_RETRIES',
+                                default=str(http.DEFAULT_MAX_RETRIES)))
+
+        parser.add_argument('--retry-interval', type=int,
+                            help='Amount of time (in seconds) between retries '
+                            'in case of conflict error (HTTP 409). '
+                            'Defaults to env[IRONIC_RETRY_INTERVAL] or %d.'
+                            % http.DEFAULT_RETRY_INTERVAL,
+                            default=cliutils.env(
+                                'IRONIC_RETRY_INTERVAL',
+                                default=str(http.DEFAULT_RETRY_INTERVAL)))
+
         # FIXME(gyee): this method should come from python-keystoneclient.
         # Will refactor this code once it is available.
         # https://bugs.launchpad.net/python-keystoneclient/+bug/1332337
@@ -236,8 +270,8 @@ class IronicShell(object):
             httplib2.debuglevel = 1
         else:
             logging.basicConfig(
-                    format="%(levelname)s %(message)s",
-                    level=logging.CRITICAL)
+                format="%(levelname)s %(message)s",
+                level=logging.CRITICAL)
 
     def do_bash_completion(self):
         """Prints all of the commands and options for bash-completion."""
@@ -292,10 +326,11 @@ class IronicShell(object):
         auth_token = kwargs.pop('auth_token', None)
         if auth_token:
             return v2_auth.Token(v2_auth_url, auth_token,
-                tenant_id=kwargs.pop('project_id', None),
-                tenant_name=kwargs.pop('project_name', None))
+                                 tenant_id=kwargs.pop('project_id', None),
+                                 tenant_name=kwargs.pop('project_name', None))
         else:
-            return v2_auth.Password(v2_auth_url,
+            return v2_auth.Password(
+                v2_auth_url,
                 username=kwargs.pop('username', None),
                 password=kwargs.pop('password', None),
                 tenant_id=kwargs.pop('project_id', None),
@@ -333,12 +368,41 @@ class IronicShell(object):
             # support only v2
             auth = self._get_keystone_v2_auth(v2_auth_url, **kwargs)
         else:
-            raise exc.CommandError('Unable to determine the Keystone version '
-                                   'to authenticate with using the given '
-                                   'auth_url.')
+            msg = _('Unable to determine the Keystone version '
+                    'to authenticate with using the given '
+                    'auth_url.')
+            raise exc.CommandError(msg)
 
         return auth
 
+    def _check_version(self, api_version):
+        if api_version == 'latest':
+            return LATEST_API_VERSION
+        else:
+            try:
+                versions = tuple(int(i) for i in api_version.split('.'))
+            except ValueError:
+                versions = ()
+            if len(versions) == 1:
+                # Default value of ironic_api_version is '1'.
+                # If user not specify the value of api version, not passing
+                # headers at all.
+                os_ironic_api_version = None
+            elif len(versions) == 2:
+                os_ironic_api_version = api_version
+                # In the case of '1.0'
+                if versions[1] == 0:
+                    os_ironic_api_version = None
+            else:
+                msg = _("The requested API version %(ver)s is an unexpected "
+                        "format. Acceptable formats are 'X', 'X.Y', or the "
+                        "literal string '%(latest)s'."
+                        ) % {'ver': api_version, 'latest': 'latest'}
+                raise exc.CommandError(msg)
+
+            api_major_version = versions[0]
+            return (api_major_version, os_ironic_api_version)
+
     def main(self, argv):
         # Parse args once to find version
         parser = self.get_base_parser()
@@ -346,8 +410,10 @@ class IronicShell(object):
         self._setup_debugging(options.debug)
 
         # build available subcommands based on version
-        api_version = options.ironic_api_version
-        subcommand_parser = self.get_subcommand_parser(api_version)
+        (api_major_version, os_ironic_api_version) = (
+            self._check_version(options.ironic_api_version))
+
+        subcommand_parser = self.get_subcommand_parser(api_major_version)
         self.parser = subcommand_parser
 
         # Handle top-level --help/-h before attempting to parse
@@ -367,7 +433,7 @@ class IronicShell(object):
             self.do_bash_completion()
             return 0
 
-        if not (args.os_auth_token and args.ironic_url):
+        if not (args.os_auth_token and (args.ironic_url or args.os_endpoint)):
             if not args.os_username:
                 raise exc.CommandError(_("You must provide a username via "
                                          "either --os-username or via "
@@ -392,22 +458,23 @@ class IronicShell(object):
 
             if not (args.os_tenant_id or args.os_tenant_name or
                     args.os_project_id or args.os_project_name):
-                raise exc.CommandError(_("You must provide a project name or"
-                    " project id via --os-project-name, --os-project-id,"
-                    " env[OS_PROJECT_ID] or env[OS_PROJECT_NAME].  You may"
-                    " use os-project and os-tenant interchangeably."))
+                raise exc.CommandError(
+                    _("You must provide a project name or"
+                      " project id via --os-project-name, --os-project-id,"
+                      " env[OS_PROJECT_ID] or env[OS_PROJECT_NAME].  You may"
+                      " use os-project and os-tenant interchangeably."))
 
             if not args.os_auth_url:
                 raise exc.CommandError(_("You must provide an auth url via "
                                          "either --os-auth-url or via "
                                          "env[OS_AUTH_URL]"))
 
-        endpoint = args.ironic_url
+        endpoint = args.ironic_url or args.os_endpoint
         service_type = args.os_service_type or 'baremetal'
         project_id = args.os_project_id or args.os_tenant_id
         project_name = args.os_project_name or args.os_tenant_name
 
-        if args.os_auth_token and args.ironic_url:
+        if (args.os_auth_token and (args.ironic_url or args.os_endpoint)):
             kwargs = {
                 'token': args.os_auth_token,
                 'insecure': args.insecure,
@@ -418,9 +485,9 @@ class IronicShell(object):
                 'auth_ref': None,
             }
         elif (args.os_username and
-            args.os_password and
-            args.os_auth_url and
-            (project_id or project_name)):
+              args.os_password and
+              args.os_auth_url and
+              (project_id or project_name)):
 
             keystone_session = kssession.Session.load_from_cli_options(args)
 
@@ -456,7 +523,16 @@ class IronicShell(object):
                 'username': args.os_username,
                 'password': args.os_password,
             }
-        client = iroclient.Client(api_version, endpoint, **kwargs)
+        kwargs['os_ironic_api_version'] = os_ironic_api_version
+        if args.max_retries < 0:
+            raise exc.CommandError(_("You must provide value >= 0 for "
+                                     "--max-retries"))
+        if args.retry_interval < 1:
+            raise exc.CommandError(_("You must provide value >= 1 for "
+                                     "--retry-interval"))
+        kwargs['max_retries'] = args.max_retries
+        kwargs['retry_interval'] = args.retry_interval
+        client = iroclient.Client(api_major_version, endpoint, **kwargs)
 
         try:
             args.func(client, args)
@@ -491,7 +567,7 @@ def main():
         print("... terminating ironic client", file=sys.stderr)
         sys.exit(130)
     except Exception as e:
-        print(str(e), file=sys.stderr)
+        print(encodeutils.safe_encode(six.text_type(e)), file=sys.stderr)
         sys.exit(1)
 
 if __name__ == "__main__":
